# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# Original framework from Netrender
# Original Octane script from Enrico Cerica

from math import atan, pi, degrees
import subprocess
import os
import sys
import time
import mathutils
import bpy
#import octanerender
import octanerender.settings
from octanerender.utils import *

class OctaneRenderEngine(bpy.types.RenderEngine):
	bl_idname = 'OCT_RENDER'
	bl_label = "Octane Render"
	bl_postprocess = False

	def _octexport(self, scene,  basePath, baseName, frame):
		# Depends on Blender version used
		try:
			from Mathutils import Vector
		except ImportError:
			from mathutils import Vector
		from itertools import zip_longest

		ff = lambda f:format(f,'.6f').rstrip('0')

		octscene = scene.octane_render
		log ('Begin export')
		# Force an update to object matrices when rendering animations
		
		sdir = '{}/{}'.format(basePath,baseName)
		export_objects=[]
		objname = '{}/{}.obj'.format(basePath,baseName)
		baseNameN=baseName+'-%s' % (frame)
		baseNameP=baseName+'-%s' % (frame-1)
		objnameN = '{}/{}.obj'.format(basePath,baseNameN)
		mtlnameP = '{}/{}.mtl'.format(basePath,baseNameP)
		
		if os.path.isfile(objnameN):
			log ('Previously created, skip ' +objnameN)
			os.remove(objname)
			os.rename(objnameN,objname)
			if os.path.isfile(mtlnameP):
				os.remove(mtlnameP)
		else:
			if octscene.export_sel_only:
				objects = bpy.context.selected_objects
			else:
				objects = scene.objects

			for obj in objects:
				# XXX LZ
				if not obj.hide_render and not obj.hide and obj.type in ['MESH', 'SURFACE', 'META', 'TEXT', 'CURVE', 'EMPTY'] and scene.layers[next((i for i in range(len(obj.layers)) if obj.layers[i]))]:
					log ('Object: {}'.format(obj.name))
					export_objects.append(obj)

			write( objname, export_objects , scene,
					octscene.export_tri,
					octscene.export_edges,
					octscene.export_normals,
					octscene.export_HQ,
					octscene.export_UV,
					octscene.export_materials,
					octscene.export_copy_images,
					octscene.export_apply_modifiers,
					octscene.export_ROTX90,
					True,False,True,False,
					octscene.export_polygroups,
					octscene.export_curves_as_nurbs)


	def render(self, scene):
		octscene = scene.octane_render
				
		if octscene.project_name=="":
			error("A project name should be defined")		
		
		log ('Plug-in starting...')
		self.update_stats("", "Octane: Exporting octscene, please wait...")
	
		unitFactor = 1
		#Set the unit factor
		unitFactor = {
			0:0.001,
			1:0.01,
			2:0.1,
			3:1,
			4:10,
			5:100,
			6:1000,
			7:0.0254,
			8:0.3048,
			9:0.9144,
			10:201.168,
			11:1609.344               
		}[int(octscene.unit_size)]

		log ('Unit Factor : %d' % (unitFactor))
	
		basePath = octscene.path.replace("\\", "/")
		imagePath = octscene.image_output.replace("\\", "/")
	
		if not os.path.exists(basePath):
			log ('Cannot find ' + basePath)
			error("Cannot find scenes directory")
	
		# Get octscene name
		baseName = octscene.project_name
		
		if os.path.isfile(octscene.project_name):
			baseName = os.path.basename(octscene.project_name)
			
		baseName = baseName.split('/')[-1].split('\\')[-1]
		
		if not baseName:
			error("Invalid octscene filename")
			
		baseName = meshNode = os.path.splitext(baseName)[0].replace(" ", "_")
	
		ocsFile = '{}/{}.ocs'.format(basePath,baseName)
		objFile = '{}/{}.obj'.format(basePath,baseName)
		# XXX
		log ('export_mode: %s' % (octanerender.export_mode))
		samplesPerImage=''
		if octanerender.export_mode=='anim':
			samplesPerImage='-s {}'.format(octscene.export_samples_per_image)
			log ('Samples per image: %d' % (octscene.export_samples_per_image))

		while True:
			log ('Rendering frame: %d' % (scene.frame_current))
			self._octexport(scene,  basePath, baseName, scene.frame_current)
			
			if octanerender.export_mode=='export':
				break
		
			lensAperture = ''
			if octscene.lens_aperture:
				lensAperture=' --cam-aperture {}'.format(octscene.lens_aperture_value)
				log ('Octane: Lens aperture: %d' % (octscene.lens_aperture_value))

# 			fdist = (octscene.camera.location-octscene.camera.data.dof_object.location).magnitude if octscene.camera.data.dof_object else octscene.camera.data.dof_distance
	
			focalDepth = ''
			if octscene.focal_depth:
				focalDepth=' --cam-focaldepth {} '.format(octscene.focal_depth_value)
				log ('Focal depth: %d' % (octscene.focal_depth_value))
				
			pngfile = '{}/{}_{:04d}.png'.format(imagePath,baseName,scene.frame_current)
		
			prefix=samplesPerImage+lensAperture+focalDepth
			
			camera_param = ''
			light_param=''
			
			axes = ['x','y','z']
			paramDsp=""
		
			
			if octscene.export_camera:
				displayError=False
				try:
					camera = bpy.data.objects[octscene.active_camera]
				except:
					displayError=True
				
				if octscene.active_camera=="" or displayError:
					error("Invalid camera object or camera not selected")
				else:
					log ('Using camera : %s' % (camera.name))
					matrix = camera.matrix_world
			
					position = rotate90x(matrix_vect(matrix,[0.0,0.0,0.0,1.0]))
					target = rotate90x(matrix_vect(matrix,[0.0,0.0,-1.0,1.0]))
					up = rotate90x(matrix_vect(matrix,[0.0,1.0,0.0,0.0]))
					for i in range(3):
						camera_param += " --cam-pos-%s %f" % (axes[i],position[i]*unitFactor)
						if i==0:paramDsp=""
						paramDsp+= "%s %f " % (axes[i],position[i]*unitFactor)
					log ('Camera position: %s' % (paramDsp))
					for i in range(3):
						camera_param += " --cam-target-%s %f" % (axes[i],target[i]*unitFactor)
						if i==0:paramDsp=""
						paramDsp+= "%s %f " % (axes[i],target[i]*unitFactor)
					log ('Camera target: %s' % (paramDsp))
					for i in range(3):
						camera_param += " --cam-up-%s %f" % (axes[i],up[i]*unitFactor)
						if i==0:paramDsp=""
						paramDsp+= "%s %f " % (axes[i],up[i]*unitFactor)
					log ('Camera up: %s' % (paramDsp))
		
					fov=camera.data.angle*180.0/3.1415926536
					camera_param += " --cam-fov %f" % fov
					log ('Octane: Camera FOV: %f' % (fov))
				
					if octscene.camera_motion:
						currentFrame=scene.frame_current
									    
						if octscene.interpolate_frame=="0":#If interpolate on next frame
							octscene.frame_set(currentFrame + 1)
							log ('Motion interploate on next frame')
						elif currentFrame>1:#Interpolate on previous frame
							octscene.frame_set(currentFrame - 1)
							log ('Motion interpolate on previous frame')
						
		
						matrix = octscene.camera.matrix_world
			
						position = rotate90x(matrix_vect(matrix,[0.0,0.0,0.0,1.0]))
						target = rotate90x(matrix_vect(matrix,[0.0,0.0,-1.0,1.0]))
						up = rotate90x(matrix_vect(matrix,[0.0,1.0,0.0,0.0]))
			
						for i in range(3):
							camera_param += " --cam-motion-pos-%s %f" % (axes[i],position[i]*unitFactor)
							if i==0:paramDsp=""
							paramDsp+= "%s %f " % (axes[i],position[i]*unitFactor)
						log ('Camera motion position: %s' % (paramDsp))
						for i in range(3):
							camera_param += " --cam-motion-target-%s %f" % (axes[i],target[i]*unitFactor)
							if i==0:paramDsp=""
							paramDsp+= "%s %f " % (axes[i],target[i]*unitFactor)
						log ('Camera motion target: %s' % (paramDsp))
						for i in range(3):
							camera_param += " --cam-motion-up-%s %f" % (axes[i],up[i]*unitFactor)
							if i==0:paramDsp=""
							paramDsp+= "%s %f " % (axes[i],up[i]*unitFactor)
						log ('Camera motion up: %s' % (paramDsp))
							
						octscene.frame_set(currentFrame)
				
			#Set the daylight option
			if octscene.export_sundir:
				displayError=False
				try:
					light = bpy.data.objects[octscene.active_light_source]
				except:
					displayError=True
				
				if octscene.active_light_source=="" or displayError:
					error("Invalid sun light object or sun light not selected")
				else:
					log ('Using light : %s' % (light.name))
					lightdata = light.data
					matrix = mathutils.Matrix(light.matrix_world.copy()).invert()
					matrix = rotate90x(matrix)
			
					for i in range(3):
						light_param += " --daylight-sundir-%s %f" % (axes[i],matrix[i][2]*unitFactor)
						if i==0:paramDsp=""
						paramDsp+= "%s %f " % (axes[i],matrix[i][2]*unitFactor)
					log ('Sun direction: %s' % (paramDsp))
	            
	#         octanePath=octaneRenderBinary.val
	#         octanePath='\"'+octanePath+'\"'
			
			objPath=os.path.join(octscene.path,objFile).replace("\\", "/")

			# Path where Octane is located
			if os.path.isdir(octscene.binary):
				octanePath = basePath+'/octane'.replace("\\", "/")
			elif os.path.isfile(octscene.binary):
				octanePath = octscene.binary.replace("\\", "/")
			else:
				error("Full path to Octane executable required")
			if not os.path.exists(octanePath):
				error("Cannot find Octane executable at specified path")
					
			projectFileAlreadyExists=False
			
			if os.path.exists(ocsFile):
					projectFileAlreadyExists=True
		
			#If linux env add quotes in file names to allow spaces
		#		if sys.platform == "linux2" or sys.platform == "darwin":
		#				objPath='\"'+objPath+'\"'
		#				objFile='\"'+objFile+'\"'
		#				ocsFile='\"'+ocsFile+'\"'
		#				pngfile='\"'+pngfile+'\"'
		
			#Set the exit after rendering flag (for animation)
			#Also set the system flag to force blender to wait Octane has finished the rendering or to immediately
			#render the control to blender (for single frame rendering)
			exitFlag=''
			commandFlag=''
			if octanerender.export_mode=='anim':
					exitFlag=' -e -q'
					#exitFlag=' -e'
		
		#				if sys.platform == "win32":
		#						commandFlag=' /wait /normal \"\" '
		#				elif sys.platform == "linux2" or sys.platform == "darwin":
		#						commandFlag = "(nice -n %d %s)"
		#		else:
		#				if sys.platform == "win32":
		#						commandFlag=' /b /normal \"\" '
		
		#				elif sys.platform == "linux2" or sys.platform == "darwin":
		#						commandFlag = "(nice -n %d %s)&"
		
			#Set the resolution option
			resolution=''
			x=octscene.resolution_width
			y=octscene.resolution_height
			if  octscene.resolution:
				resSize = octscene.percent_size/100
				log ('Resolution factor : %d' % (resSize))
				resolution=' --film-width %i --film-height %i' % (x*resSize,y*resSize)
		
			#Set the GPUs option
			useGPU=''
			if octscene.GPU_selector:
				for val in octscene.GPU_use_list.split(' '):
						useGPU+=' -g '+val
	
			command =  octanePath+exitFlag+resolution+useGPU
			command =  exitFlag+resolution+useGPU
			
			#Use an existing octane ocs file
			if projectFileAlreadyExists and not octscene.replace_project:
				if octanerender.export_mode=='anim':
					# XXX LZ
					if octscene.relink_obj == True:
						command+=' -m %s.obj -r %s %s --output-png %s %s %s %s' % (meshNode,objFile,prefix,pngfile,camera_param,light_param,ocsFile)
					else:
						command+=' -m %s.obj %s --output-png %s %s %s %s' % (meshNode,prefix,pngfile,camera_param,light_param,ocsFile)
				else:
					if octscene.relink_obj == True:
						command+=' -m %s.obj -r %s %s %s %s %s' % (meshNode,objFile,prefix,camera_param,light_param,ocsFile)
					else:
						command+=' -m %s.obj %s %s %s %s' % (meshNode,prefix,camera_param,light_param,ocsFile)
		
			#Force creation of the octane ocs file if it doesn't exist
			else:
				log ('Creating or replacing existing project')
					
				if octanerender.export_mode=='anim':
					command+=' -n %s -m %s.obj -l %s %s --output-png %s %s %s' % (ocsFile,meshNode,objPath,prefix,pngfile,camera_param,light_param)
				else:
					command+=' -n %s -m %s.obj -l %s %s %s %s' % (ocsFile,meshNode,objPath,prefix,camera_param,light_param)
		
				#Reset the replace project flag to avoid accidental replacement
				#bpy.context.scene.octane_render.replace_project=False
		
		
			#If no batch file should be created -> start octane immediately
		#		else:
		#				if ostype == "win32":
		#						command='start ' + commandFlag + command
		#						log (command)
		#						os.system(command)
		
		#				elif ostype == "linux2" or ostype == "darwin":
		#						if exitAfter:
		#								command=commandFlag%(0, command)
		#								log (command)
		#								subprocess.call(command,shell=True)
		#						else:
		#								command=commandFlag%(0, command)
		#								log (command)
		#								os.system(command)
		
		#		log('launch Octane: {} {}/{}.ocs'.format(exepath,basePath,baseName))
			log ('Launch Octane: {} {} '.format(octanePath,command))
		#		octproc = subprocess.Popen([octanePath,'{}'.format(command)], cwd=basePath, shell=False)
			if sys.platform == "win32":
				octproc = subprocess.Popen('{} {}'.format(octanePath,command))
			else:
				octproc = subprocess.Popen([octanePath,'{}'.format(command)])
		
			if octanerender.export_mode != 'anim':
				break
				
			# Wait for OCT , convert and load image
			self.update_stats("", "Octane: Batch Rendering frame# {} (see console for progress), please wait...".format(scene.frame_current))
			# Hold the render results window hostage until OCT returns...
			log ('Finished frame: %d' % (scene.frame_current))
			bpy.context.scene.frame_set(scene.frame_current+octscene.animation_step)
			if scene.frame_current > octscene.animation_end:
				break
			log ('Next frame: %d' % (scene.frame_current))
			while True:
				if octproc.poll() != None:
					result = self.begin_result(0, 0, x, y)
					try:
						log ('Load image from file: {}'.format(pngfile))
						result.layers[0].load_from_file('{}'.format(pngfile))
					except:
						pass

					self.end_result(result)
					break
					
				if self.test_break():
					try:
						octproc.terminate()
					except:
						pass
					break
					
				time.sleep(1)


