# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# Original framework from Netrender
# Original Octane script from Enrico Cerica

from math import atan, pi, degrees
import subprocess
import os
import sys
import time
import mathutils
import bpy
#import octanerender
import octanerender.settings

from octanerender.utils import *

ostype = sys.platform
PATH_PREFIX = "/tmp/"

def base_poll(cls, context):
	rd = context.scene.render
	return (rd.use_game_engine==False) and (rd.engine in cls.COMPAT_ENGINES)

def octane_file():
	if octanerender.octane_file != bpy.data.filepath:
		octanerender.octane_file = bpy.data.filepath
		octanerender.octane_data = True

def octane_data(octanesettings):
	octane_file()

	if octanerender.octane_data:
		octanerender.octane_data = False

# Main Octane Panel
class OctaneRenderButtonsPanel():
	bl_space_type = "PROPERTIES"
	bl_region_type = "WINDOW"
	bl_context = "render"
	# COMPAT_ENGINES must be defined in each subclass, external engines can add themselves here

	@classmethod
	def poll(cls, context):
		rd = context.scene.render
		return rd.engine == 'OCT_RENDER' and rd.use_game_engine == False 

# Octane Render Plugin Panel
class RENDER_OCT_render_settings(OctaneRenderButtonsPanel, bpy.types.Panel):
	bl_label = "Octane Render Plug-in " + octanerender.Version
	COMPAT_ENGINES = {'OCT_RENDER'}    

	@classmethod
	def poll(cls, context):
		return super().poll(context)

	def draw_header(self, context):
		layout = self.layout
		layout.label(text="", icon='SEQUENCE')

	def draw(self, context):
		layout = self.layout

		scene = context.scene
		octanesettings = scene.octane_render

		layout.prop(octanesettings, "panel_mode", expand=True)

		#layout.separator()

		if octanesettings.panel_mode == "MODE_EXPORT":
			layout.operator("button_export",    icon='SCENE_DATA')
		elif octanesettings.panel_mode == "MODE_RENDER":
			layout.operator("button_render",    icon='RENDER_STILL')
		else:
			layout.operator("button_animation", icon='RENDER_ANIMATION')

			row = layout.row()
			col = row.column()
			col.prop(octanesettings, "image_output")

			row = layout.row()
			col = row.column()
			col.prop(octanesettings, "export_samples_per_image")
			col = row.column()
			col.prop(octanesettings, "export_samples_presets")

# Dimensions Panel
class RENDER_OCT_dimensions(OctaneRenderButtonsPanel, bpy.types.Panel):
	bl_label = "Dimensions"
	COMPAT_ENGINES = {'OCT_RENDER'}

	def draw_header(self, context):
		layout = self.layout
		layout.label(text="", icon='SEQUENCE')

	def draw(self, context):
		layout = self.layout

		scene = context.scene
		octanesettings = scene.octane_render
		# Initialise resolution
		if octanesettings.resolution_width==0 or octanesettings.resolution_height==0:
			r = scene.render
			octanesettings.resolution_width = int(r.resolution_x)
			octanesettings.resolution_height = int(r.resolution_y)

		row = layout.row(align=True)
		row.menu("RENDER_MT_presets", text=bpy.types.RENDER_MT_presets.bl_label)
		#row.operator("render.preset_add", text="", icon="ZOOMIN")
		#row.operator("render.preset_add", text="", icon="ZOOMOUT").remove_active = True

		split = layout.split()

		col = split.column()
		sub = col.column(align=True)
		sub.label(text="Resolution:")
		sub.prop(octanesettings, "resolution_width", text="X")
		sub.prop(octanesettings, "resolution_height", text="Y")
		sub.prop(octanesettings, "percent_size", text="")

		col = split.column()
		sub = col.column(align=True)
		sub.label(text="Frame Range:")
		sub.prop(octanesettings, "animation_start", text="Start")
		sub.prop(octanesettings, "animation_end", text="End")
		sub.prop(octanesettings, "animation_step", text="Step")

# Octane Project Settings Panel
class RENDER_OCT_project_settings(OctaneRenderButtonsPanel, bpy.types.Panel):
	bl_label = "Octane Project Settings"
	COMPAT_ENGINES = {'OCT_RENDER'}    

	@classmethod
	def poll(cls, context):
		return super().poll(context)

	def draw_header(self, context):
		layout = self.layout
		layout.label(text="", icon='URL')

	def draw(self, context):
		layout = self.layout

		scene = context.scene
		octanesettings = scene.octane_render

		if octanesettings.path == "":
			layout.prop(octanesettings, "path",icon='ERROR')
		else:
			layout.prop(octanesettings, "path")

		if octanesettings.project_name == "":
			layout.prop(octanesettings, "project_name",icon='ERROR')
		else:
			layout.prop(octanesettings, "project_name")
		layout.prop(octanesettings, "unit_size")

# Octane Environment Settings Panel
class RENDER_OCT_environment_settings(OctaneRenderButtonsPanel, bpy.types.Panel):
	bl_label = "Octane Environment Settings"
	#bl_options = {'DEFAULT_CLOSED'}
	COMPAT_ENGINES = {'OCT_RENDER'}    

	@classmethod
	def poll(cls, context):
		return super().poll(context)

	def draw_header(self, context):
		layout = self.layout
		layout.label(text="", icon='WORLD')

	def draw(self, context):
		layout = self.layout

		scene = context.scene
		octanesettings = scene.octane_render

		layout.prop(octanesettings, "replace_project")
		if octanesettings.replace_project == False:
			layout.prop(octanesettings, "relink_obj")

		# Initialise with active or first camera found
		#if ocscene.active_camera=='':
		#	ocscene.oct_active_camera=ocscene.camera.name
			
		# Initialise the source light as the first sun light found
		#if octanesettings.active_light_source=='':	
		#	for lamp in [l for l in scene.objects if l.type == 'LAMP']:
		#		if lamp.data.type=="SUN":
		#			octanesettings.active_light_source=lamp.name
		#			break

		# Camera
		layout.separator()
		row = layout.row()
		row.label(text="Camera Properties", icon='SCENE')
		layout.prop(octanesettings, "export_camera")
		if octanesettings.export_camera:
			row = layout.row()
			split = row.split(percentage=0.75)
			col = split.column()
			if octanesettings.active_camera == "":
				col.prop(octanesettings, "active_camera",icon='ERROR')
			else:
				col.prop(octanesettings, "active_camera")
			col = split.column()
			col.operator("button_select_camera",icon='SCENE')

			#row = layout.row()
			layout.prop(octanesettings, "lens_aperture")
			if octanesettings.lens_aperture:
				row = layout.row()
				split = row.split(percentage=0.75)
				col = split.column()
				col.prop(octanesettings, "lens_aperture_value")
				col = split.column()
				col.prop(octanesettings, "lens_aperture_curve")

			#row = layout.row()
			layout.prop(octanesettings, "focal_depth")
			if octanesettings.focal_depth:
				row = layout.row()
				split = row.split(percentage=0.75)
				col = split.column()
				col.prop(octanesettings, "focal_depth_value")
				col = split.column()
				col.prop(octanesettings, "focal_depth_curve")

			#row = layout.row()
			layout.prop(octanesettings, "camera_motion")
			if octanesettings.camera_motion:
				row = layout.row()
				split = row.split(percentage=0.75)
				col = split.column()
				col.prop(octanesettings, "interpolate_frame")
			
		# Day light environment
		layout.separator()
		row = layout.row()
		row.label(text="Daylight Environment", icon='LAMP_SUN')
		layout.prop(octanesettings, "export_sundir")
		if octanesettings.export_sundir:
			row = layout.row()
			split = row.split(percentage=0.75)
			col = split.column()
			if octanesettings.active_light_source == "":
				col.prop(octanesettings, "active_light_source",icon='ERROR')
			else:
				col.prop(octanesettings, "active_light_source")
			col = split.column()
			col.operator("button_select_light_source", icon='LAMP_SUN')

# Octane Export Settings
class RENDER_OCT_export_settings(OctaneRenderButtonsPanel, bpy.types.Panel):
	bl_label = "Octane Export Settings"
	bl_options = {'DEFAULT_CLOSED'}
	COMPAT_ENGINES = {'OCT_RENDER'}    

	@classmethod
	def poll(cls, context):
		return super().poll(context)

	def draw_header(self, context):
		layout = self.layout
		layout.label(text="", icon='SCRIPTWIN')

	def draw(self, context):
		layout = self.layout

		scene = context.scene
		octanesettings = scene.octane_render

		# Context
		row = layout.row()
		row.label(text=" Context",icon='GROUP')
		row = layout.row()
		split = row.split(percentage=0.50)
		col = split.column()
		col.prop(octanesettings, "export_sel_only")
		col = split.column()
		col.prop(octanesettings, "export_remove_hidden")
		
		# Output options
		layout.separator()
		row = layout.row()
		row.label(text=" Output Options",icon='MOD_MESHDEFORM')
		row = layout.row()
		split = row.split(percentage=0.33)

		col = split.column()
		col.prop(octanesettings, "export_apply_modifiers")
		col = split.column()
		col.prop(octanesettings, "export_ROTX90")
		col = split.column()
		col.prop(octanesettings, "export_copy_images")

		layout.separator()
		row = layout.row()
		row.label(text=" Export Options",icon='SETTINGS')

		row = layout.row()
		split = row.split(percentage=0.25)
		col = split.column()
		col.prop(octanesettings, "export_edges")
		col = split.column()
		col.prop(octanesettings, "export_tri")
		col = split.column()
		col.prop(octanesettings, "export_materials")
		col = split.column()
		col.prop(octanesettings, "export_UV")

		row = layout.row()
		split = row.split(percentage=0.25)
		col = split.column()
		col.prop(octanesettings, "export_normals")
		col = split.column()
		col.prop(octanesettings, "export_HQ")
		col = split.column()
		col.prop(octanesettings, "export_polygroups")
		col = split.column()
		col.prop(octanesettings, "export_curves_as_nurbs")

# Octane system Settings
class RENDER_OCT_system_settings(OctaneRenderButtonsPanel, bpy.types.Panel):
	bl_label = "Octane System Settings"
	COMPAT_ENGINES = {'OCT_RENDER'}    

	@classmethod
	def poll(cls, context):
		return super().poll(context)

	def draw_header(self, context):
		layout = self.layout
		layout.label(text="", icon='RADIO')

	def draw(self, context):
		layout = self.layout
		scene = context.scene
		octanesettings = scene.octane_render
		if octanesettings.binary == "":
			layout.prop(octanesettings, "binary",icon='ERROR')
		else:
			layout.prop(octanesettings, "binary")

		row = layout.row()
		split = row.split(percentage=0.5)
		col = split.column()
		col.prop(octanesettings, "GPU_selector")
		if octanesettings.GPU_selector:
			col = split.column()
			col.prop(octanesettings, "GPU_use_list")
		layout.prop(octanesettings, "resolution")
		layout.prop(octanesettings, "verbose")
		if octanesettings.verbose == True:
			if octanerender.Verbose == False:
				octanerender.Verbose = True
				log('Turning verbose mode on')
			octanerender.Verbose = True
		else:
			if octanerender.Verbose == True:
				log('Turning verbose mode off')
			octanerender.Verbose = False


